fromList
  [ ( "中文"
    , fromList
        [ ("鸡丁", "宫保")
        , ("豆腐", "麻婆")
        ]
    )
  , ( "русский"
    , fromList [ ( "хорошо", "очень" ) ]
    )
  , ( "العَرَبِيَّة‎‎"
    , fromList
        [ ("واحد", "١")
        , ("اثنان", "٢")
        , ("ثلاثة", "٣")
        ]
    )
  ]
